@@ -259,4 +259,109 @@
 					/* 254 for set_thread_area */
 					/* 255 for get_thread_area */
 					/* 256 for set_tid_address */
-#define TARGET_NR_utimes		        (269)
+#define TARGET_NR_set_tid_address		256
+#define TARGET_NR_timer_create		257
+#define TARGET_NR_timer_settime		258
+#define TARGET_NR_timer_gettime		259
+#define TARGET_NR_timer_getoverrun		260
+#define TARGET_NR_timer_delete		261
+#define TARGET_NR_clock_settime		262
+#define TARGET_NR_clock_gettime		263
+#define TARGET_NR_clock_getres		264
+#define TARGET_NR_clock_nanosleep		265
+#define TARGET_NR_statfs64			266
+#define TARGET_NR_fstatfs64			267
+#define TARGET_NR_tgkill			268
+#define TARGET_NR_utimes			269
+#define TARGET_NR_arm_fadvise64_64		270
+#define TARGET_NR_pciconfig_iobase		271
+#define TARGET_NR_pciconfig_read		272
+#define TARGET_NR_pciconfig_write		273
+#define TARGET_NR_mq_open			274
+#define TARGET_NR_mq_unlink			275
+#define TARGET_NR_mq_timedsend		276
+#define TARGET_NR_mq_timedreceive		277
+#define TARGET_NR_mq_notify			278
+#define TARGET_NR_mq_getsetattr		279
+#define TARGET_NR_waitid			280
+#define TARGET_NR_socket			281
+#define TARGET_NR_bind			282
+#define TARGET_NR_connect			283
+#define TARGET_NR_listen			284
+#define TARGET_NR_accept			285
+#define TARGET_NR_getsockname		286
+#define TARGET_NR_getpeername		287
+#define TARGET_NR_socketpair			288
+#define TARGET_NR_send			289
+#define TARGET_NR_sendto			290
+#define TARGET_NR_recv			291
+#define TARGET_NR_recvfrom			292
+#define TARGET_NR_shutdown			293
+#define TARGET_NR_setsockopt			294
+#define TARGET_NR_getsockopt			295
+#define TARGET_NR_sendmsg			296
+#define TARGET_NR_recvmsg			297
+#define TARGET_NR_semop			298
+#define TARGET_NR_semget			299
+#define TARGET_NR_semctl			300
+#define TARGET_NR_msgsnd			301
+#define TARGET_NR_msgrcv			302
+#define TARGET_NR_msgget			303
+#define TARGET_NR_msgctl			304
+#define TARGET_NR_shmat			305
+#define TARGET_NR_shmdt			306
+#define TARGET_NR_shmget			307
+#define TARGET_NR_shmctl			308
+#define TARGET_NR_add_key			309
+#define TARGET_NR_request_key		310
+#define TARGET_NR_keyctl			311
+#define TARGET_NR_semtimedop			312
+#define TARGET_NR_vserver			313
+#define TARGET_NR_ioprio_set			314
+#define TARGET_NR_ioprio_get			315
+#define TARGET_NR_inotify_init		316
+#define TARGET_NR_inotify_add_watch		317
+#define TARGET_NR_inotify_rm_watch		318
+#define TARGET_NR_mbind			319
+#define TARGET_NR_get_mempolicy		320
+#define TARGET_NR_set_mempolicy		321
+#define TARGET_NR_openat			(322)
+#define TARGET_NR_mkdirat			(323)
+#define TARGET_NR_mknodat			(324)
+#define TARGET_NR_fchownat			(325)
+#define TARGET_NR_futimesat			(326)
+#define TARGET_NR_fstatat64			(327)
+#define TARGET_NR_unlinkat			(328)
+#define TARGET_NR_renameat			(329)
+#define TARGET_NR_linkat			(330)
+#define TARGET_NR_symlinkat			(331)
+#define TARGET_NR_readlinkat			(332)
+#define TARGET_NR_fchmodat			(333)
+#define TARGET_NR_faccessat			(334)
+					/* 335 for pselect6 */
+					/* 336 for ppoll */
+#define TARGET_NR_unshare			(337)
+#define TARGET_NR_set_robust_list		(338)
+#define TARGET_NR_get_robust_list		(339)
+#define TARGET_NR_splice			(340)
+#define TARGET_NR_arm_sync_file_range	(341)
+#define TARGET_NR_sync_file_range2		TARGET_NR_arm_sync_file_range
+#define TARGET_NR_tee			(342)
+#define TARGET_NR_vmsplice			(343)
+#define TARGET_NR_move_pages			(344)
+#define TARGET_NR_getcpu			(345)
+					/* 346 for epoll_pwait */
+#define TARGET_NR_kexec_load			(347)
+#define TARGET_NR_utimensat			(348)
+#define TARGET_NR_signalfd			(349)
+#define TARGET_NR_timerfd			(350)
+#define TARGET_NR_eventfd			(351)
+#define TARGET_NR_fallocate			(352)
+#define TARGET_NR_timerfd_settime		(353)
+#define TARGET_NR_timerfd_gettime		(354)
+#define TARGET_NR_signalfd4			(355)
+#define TARGET_NR_eventfd2			(356)
+#define TARGET_NR_epoll_create1		(357)
+#define TARGET_NR_dup3				(358)
+#define TARGET_NR_pipe2			(359)
+#define TARGET_NR_inotify_init1		(360)
